import sys, os, time, random#, ipdb
import numpy as np
import pandas as pd
from scipy import sparse

import warnings
warnings.filterwarnings('ignore')

def setup():
   # This function loads the simulation starting data
   users = np.load("../user_polarity.npy" % run, allow_pickle = True)     # This is a vector with a value per user: its polarity from -1 to +1
   sources = np.load("../source_polarity.npy" % run, allow_pickle = True) # This is a vector with a value per source: its polarity from -1 to +1
   social = sparse.load_npz("../social_network.npz" % run).tocsr()        # This is the social network, connecting users to their friends (homophily, communities)
   audience = sparse.load_npz("../audience_network.npz" % run).todense()  # This is the audience network, bipartite connecting users to the sources they follow
   polarity_delta = np.abs(np.subtract(np.matrix(users).T, np.matrix(sources)))         # Calculate the difference in polarity between each user, source pair
   sharers = polarity_delta < shareability                                              # Boolean: if the user sees an item from the source, will they re-share it?
   flaggers = polarity_delta > flaggability                                             # Boolean: if the user sees an item from the source, will they flag it?
   if not os.path.exists(folder):       # If the output folder doesn't exists, create it.
      os.makedirs(folder, exist_ok = True)
   return sources, users, sharers, flaggers, social, audience, polarity_delta

def cascade(sharers, flaggers, social, audience, users):
   shares = np.logical_and(audience, sharers.T)                                              # At the first step a source is reshared by users following it and passing the share test
   flags = np.logical_and(audience, flaggers.T).astype(int)                                  # At the first step a source is flagged by users following it and passing the flag test
   total_shares = shares.copy()                                                              # We keep track of the total number of shares for each source in the cascade
   total_audience = audience.copy()                                                          # We keep track of the users seeing items produced by a source in the cascade
   for i in range(audience.shape[0]):                                                        # For each source...
      while shares[i].sum() > 0:                                                             # ...if there's at least one user resharing the source's content...
         current_audience = sparse.csr_matrix(shares[i]).dot(social).todense()               # Update the current audience of the item, who are the friends of the sharers
         flags[i] += np.logical_and(current_audience, flaggers.T[i]).astype(int)             # Update the flags received by the source, who are the flaggers in the current audience
         shares[i] = np.logical_and(current_audience, sharers.T[i])                          # Update the sharers at this time step. These are the sharers in the current audience...
         shares[i] = np.logical_and(shares[i], np.logical_xor(shares[i], total_shares[i]))   # ...who did not already share the content (otherwise this would cause an infinite loop)
         total_shares[i] += shares[i]                                                        # We keep track of the total number of shares for each source in the cascade
         total_audience[i] += current_audience                                               # We keep track of the users seeing items produced by a source in the cascade
   return total_shares, flags, total_audience

def update_sources(sources, users, social, audience, flags, total_shares, total_audience):
   sources_hyp = np.array((users.dot(audience.T) / audience.sum(axis = 1).T))[0]                                     # Sources try to move their polarity by averaging their audience's polarity
   sources_hyp[sources_hyp > 1] = 1                                                                                  # If the polarity goes out of bounds, we correct it
   sources_hyp[sources_hyp < -1] = -1
   sources_hyp[np.isnan(sources_hyp)] = sources[np.isnan(sources_hyp)]                                               # If the source did not have an audience, its polarity stays where it was
   polarity_delta_hyp = np.abs(np.subtract(np.matrix(users).T, np.matrix(sources_hyp)))                              # Recalculate the polarity difference between the new sources and the users
   sharers_hyp = polarity_delta_hyp < shareability                                                                   # Boolean: if the user sees an item from the new source, will they re-share it?
   flaggers_hyp = polarity_delta_hyp > flaggability                                                                  # Boolean: if the user sees an item from the new source, will they flag it?
   total_shares_hyp, flags_hyp, total_audience_hyp = cascade(sharers_hyp, flaggers_hyp, social, audience, users)     # Perform the hypothetical cascade with the new source polarity values
   movers1 = (flags.sum(axis = 1) - flags_hyp.sum(axis = 1)) / (flags.sum(axis = 1) + 1)                             # Sources accept the new value if it generates fewer flags than the old polarity
   movers2 = (total_shares_hyp.sum(axis = 1) - total_shares.sum(axis = 1)) / (total_shares.sum(axis = 1) + 1)        # proportionally it generates more shares
   movers = np.array(((movers1 + movers2) / 2) > greediness).T[0]
   _sources = sources.copy()                                                                                         # Copy old polarity values to keep track of how much the sources moved
   sources[movers], polarity_delta[:,movers] = sources_hyp[movers], polarity_delta_hyp[:,movers]                     # Update the polarity values for the sources that moved
   return sources, polarity_delta, np.abs(sources - _sources).mean(), movers

def update_users(sources, total_audience, users, social):
   user_flaggers = np.abs(np.subtract(np.matrix(users).T, np.matrix(users))) > flaggability                         # Identify the user pairs whose polarity is so different they'd flag each other
   userflags = social.todense() & user_flaggers                                                                     # Select the social connections between users who'd flag each other
   #social_pressure_push = np.array(users.dot(userflags) / userflags.sum(axis = 0))[0]                               # The users who'd flag each other push each other away from their polarity
   social_pressure_push = np.array(users.dot(userflags))[0]
   #social_pressure_push[np.isnan(social_pressure_push)] = users[np.isnan(social_pressure_push)]                     # If the user had no flagging friend, they do not experience a push
   usernoflags = social.todense() & ~user_flaggers                                                                  # Select the social connections between users who'd NOT flag each other
   #social_pressure_pull = np.array(users.dot(usernoflags) / usernoflags.sum(axis = 0))[0]                           # The users who'd NOT flag each other update their polarity with the avg of friends
   social_pressure_pull = np.array(users.dot(usernoflags))[0]
   #social_pressure_pull[np.isnan(social_pressure_pull)] = users[np.isnan(social_pressure_pull)]                     # If the user had no attracting friend, they do not experience a pull
   #social_pressure = ((users + social_pressure_pull) + (users - social_pressure_push)) / 2                          # The final social position is the sum between social push and pull
   #ipdb.set_trace()
   social_pressure = np.array((social_pressure_pull - social_pressure_push + (users * np.array((userflags.sum(axis = 0) - usernoflags.sum(axis = 0)))[0])) / (skepticism * social.sum(axis = 0)))[0]
   social_pressure[np.isnan(social_pressure)] = 0                                    # If the user experiences push nor pull, they stay where they were
   influenced_by_media_push = total_audience.copy()                                                                 # Same procedure, but with media and audience links
   influenced_by_media_push[~flaggers.T] = 0                                                                        # For the media push, we ignore the sources a user would not flag
   #media_pressure_push = np.array(sources.dot(influenced_by_media_push) / influenced_by_media_push.sum(axis = 0))[0]
   media_pressure_push = np.array(sources.dot(influenced_by_media_push))[0]
   #media_pressure_push[np.isnan(media_pressure_push)] = users[np.isnan(media_pressure_push)]
   influenced_by_media_pull = total_audience.copy()
   influenced_by_media_pull[flaggers.T] = 0                                                                         # For the media pull, we ignore the sources a user would flag
   #media_pressure_pull = np.array(sources.dot(influenced_by_media_pull) / influenced_by_media_pull.sum(axis = 0))[0]
   media_pressure_pull = np.array(sources.dot(influenced_by_media_pull))[0]
   #media_pressure_pull[np.isnan(media_pressure_pull)] = users[np.isnan(media_pressure_pull)]
   #media_pressure = ((users + media_pressure_pull) + (users - media_pressure_push)) / 2
   media_pressure = np.array((media_pressure_pull - media_pressure_push + (users * np.array((influenced_by_media_push.sum(axis = 0) - influenced_by_media_pull.sum(axis = 0)))[0])) / (skepticism * total_audience.sum(axis = 0)))[0]
   media_pressure[np.isnan(media_pressure)] = 0
   
   _users = users + social_pressure + media_pressure
   #_users = np.nanmean([social_pressure, media_pressure], axis = 0)
   #_users = np.nansum([users, np.nanmean([social_pressure, media_pressure], axis = 0) * 1], axis = 0) / (1 + 1) # The new user position is a wavg between old position and the social-media pressure
                                                                                                                    # The parameter regulates the "stubborness" of the user. 1 = equal weight between
                                                                                                                    # old polarity and experienced pressure.
   _users[_users > 1] = 1                                                                                           # If the polarity goes out of bounds, we correct it
   _users[_users < -1] = -1
   return _users, np.abs(users - _users).mean()

def update_audience(polarity_delta, audience):
   # Note: Audience is not binary, but weighted: users satisfy their news need with the same source if there's nothing else attractive enough
   unfollows = np.array(np.array(audience, dtype = bool) & (polarity_delta.T > flaggability)) # Users identify which sources they're reading that are beyond their tolerance
   needing_new_sources = np.ma.sum(np.ma.masked_array(audience, mask = ~unfollows), axis = 0) # Calculates how many sources they are unfollowing, important to mantain their news need
   needing_new_sources = np.array(needing_new_sources.data)[0]
   audience[unfollows] = 0                                                                    # Destroys the user-source follow link
   for i in range(needing_new_sources.shape[0]):                                              # For all users...
      if needing_new_sources[i] > 0:                                                          # ...if the user needs at least one new source to follow...
         p = (1 / (polarity_delta[i] + 0.01))                                                 # The likelihood of following a new source is proportional to how close their polarity is
         new_sources = np.random.choice(range(sources.shape[0]), size = needing_new_sources[i], p = np.array(p / p.sum())[0]) # Picks new source to follow (with replacement: users can multiple follow)
         for ns in new_sources:
            audience[ns, i] += 1                                                              # Creates the new links in the user-source net and/or increases the weight of the already existing ones
   return audience

def update_social(social, users):
   user_polarity_delta = np.abs(np.subtract(np.matrix(users).T, np.matrix(users)))                         # Calculate how distante the polarity is between all pairs of users
   user_flaggers = user_polarity_delta > flaggability                                                      # Identify the user pairs beyond the flagging distance: these users would flag each other
   needing_new_friends = np.ma.sum(np.ma.masked_array(social.todense(), mask = ~user_flaggers), axis = 0)  # Identify all users who are beyond a users's flagging distance
   needing_new_friends = np.round(np.array(needing_new_friends.data)[0] * 0.51).astype(int)                # For each social connection deleted there's a 51% chance the user will make a new friend
   social[user_flaggers] = 0                                                                               # Unfriends users that are beyond the flagging threshold of a user
   social.eliminate_zeros()
   sim = np.array(audience.T.dot(audience))                                                                # User-user similarity based on the news sources they follow
   np.fill_diagonal(sim, 0)                                                                                # We don't care about the similarity of a user with themselves
   sim[social.nonzero()] = 0
   sim = sim / sim.sum(axis = 0)                                                                           # Transform the similarity into a probability
   for i in range(needing_new_friends.shape[0]):                                                           # For each user...
      if needing_new_friends[i] > 0 and not np.isnan(sim[:,i]).any():                                      # ...if they need a new friend and they follow at least a source...
         new_friends = np.random.choice(range(users.shape[0]), size = needing_new_friends[i], p = sim[:,i])# ...pick random new friends proportionally to the similarity of the sources they follow
         for nf in new_friends:                                                                            # For each new friend...
            social[nf, i] = 1                                                                              # ...create the social link
            needing_new_friends[nf] -= 1                                                                   # Keep track that the new friend needs one fewer friend than they did.
   social += social.T                                                                                      # Make sure the social network is symmetric
   social[social > 1] = 1                                                                                  # Make sure the social network is binary
   return social

def main_loop(sources, sharers, flaggers, social, audience, users, polarity_delta):
   start = time.time()
   total_shares, flags, total_audience = cascade(sharers, flaggers, social, audience, users) # Perform the first cascade with the initial conditions
   sys.stderr.write(f"Run {run} ({time.time() - start:.2f}s) (step, usrchng, usrpol, srcchng, srcpol)\t0\t0.00000\t{users.std():.3f}\t0.00000\t{sources.std():.3f}\n")
   pressure_delta = 1
   source_delta = 1
   i = 0
   with open(f"{folder}/run_{run}.log", 'w') as f:
      f.write("step\tusrchng\tusrpol\tsrcchng\tsrcpol\n")
      while (pressure_delta > 1e-3 or source_delta > 1e-3) and i < 100: # We keep going until convergence (neither users nor sources move more than 1%) or a sufficient number of iterations
         start = time.time()
         i += 1
         _users, pressure_delta = update_users(sources, total_audience, users, social) # Users modify their polarity according to their friends and the news they read
         # Sources modify their polarity according to the flags they received
         _sources, _polarity_delta, source_delta, movers = update_sources(sources, users, social, audience, flags, total_shares, total_audience)
         audience = update_audience(polarity_delta, audience) # Users unfollow news at large polarity difference and replace them with news at smaller polarity difference
         social = update_social(social, users) # Users unfriend users at large polarity difference and befriend those reading similar news as they do
         users = _users # Apply the changes to the users
         sources = _sources # Apply the changes to the sources
         polarity_delta = _polarity_delta # Update the user-source polarity difference
         total_shares, flags, total_audience = cascade(sharers, flaggers, social, audience, users) # Perform the second cascade with the new polarities, social, and audience networks
         log_line = f"{i}\t{pressure_delta:.5f}\t{users.std():.3f}\t{source_delta:.5f}\t{sources.std():.3f}\n"
         sys.stderr.write(f"Run {run} ({time.time() - start:.2f}s) (step, usrchng, usrpol, srcchng, srcpol)\t{log_line}")
         f.write(log_line)
   return total_shares, flags, audience, users, social

def write_output_sources(sources, total_shares, flags, audience):
   df = pd.DataFrame()
   df["initial_polarity"] = np.load("../04_user_mobile_model/synth_data/run%s/source_polarity.npy" % run, allow_pickle = True)
   df["final_polarity"] = sources
   df["shares"] = np.array(total_shares).sum(axis = 1)
   df["flags"] = np.array(flags).sum(axis = 1)
   df["initial_audience"] = np.array(sparse.load_npz("../04_user_mobile_model/synth_data/run%s/audience_network.npz" % run).todense()).sum(axis = 1)
   df["final_audience"] = np.array(audience).sum(axis = 1)
   df.to_csv(outfile_sources, index = False, sep = "\t")

def write_output_users(users):
   df = pd.DataFrame()
   df["user"] = range(users.shape[0])
   df["initial_polarity"] = np.load("../04_user_mobile_model/synth_data/run%s/user_polarity.npy" % run, allow_pickle = True)
   df["final_polarity"] = users
   df.to_csv(outfile_users, index = False, sep = "\t")

shareability = float(sys.argv[1])
flaggability = float(sys.argv[2])
greediness = float(sys.argv[3])
skepticism = float(sys.argv[4])
run = int(sys.argv[5])
folder = f"results/{shareability}_{flaggability}_{greediness}_{skepticism}"
outfile_sources = f"{folder}/sources_{run}.csv"
outfile_users = f"{folder}/users_{run}.csv"

if not os.path.exists(outfile_sources):
   sources, users, sharers, flaggers, social, audience, polarity_delta = setup()
   total_shares, flags, audience, users, social = main_loop(sources, sharers, flaggers, social, audience, users, polarity_delta)
   write_output_sources(sources, total_shares, flags, audience)
   write_output_users(users)
   sparse.save_npz(f"{folder}/social_{run}", social)
